"""DingTalk 适配器消息。"""
from typing import Any, Dict, List, Optional

from alicebot.message import MessageSegment

__all__ = ["DingTalkMessage"]


class DingTalkMessage(MessageSegment):  # type: ignore
    """DingTalk 消息"""

    @property
    def _message_class(self) -> None:
        return None

    def __str__(self):
        if self.type == "text":
            return self.data["content"]
        return super().__str__()

    def get_plain_text(self) -> str:
        """获取消息中的纯文本部分。

        Returns:
            消息中的纯文本部分。
        """
        if self.type == "text":
            return self.data["content"]
        return ""

    def as_dict(self) -> Dict[str, Dict[str, Any]]:
        """返回符合钉钉消息标准的消息字段字典。

        Returns:
            符合钉钉消息标准的消息字段字典。
        """
        if self.type == "raw":
            return self.data
        return {self.type: self.data}

    @classmethod
    def raw(cls, data: Dict[str, Any]) -> "DingTalkMessage":
        """DingTalk 原始消息"""
        return cls(type="raw", data=data)

    @classmethod
    def text(cls, content: str) -> "DingTalkMessage":
        """DingTalk text 消息"""
        return cls(type="text", data={"content": content})

    @classmethod
    def link(
        cls, text: str, title: str, message_url: str, pic_url: Optional[str] = None
    ):
        """DingTalk link 消息"""
        return cls(
            type="link",
            data={
                "text": text,
                "title": title,
                "messageUrl": message_url,
                "picUrl": pic_url,
            },
        )

    @classmethod
    def markdown(cls, title: str, text: str):
        """DingTalk markdown 消息"""
        return cls(type="markdown", data={"title": title, "text": text})

    @classmethod
    def action_card_single_btn(
        cls,
        title: str,
        text: str,
        single_title: str,
        single_url: str,
        btn_orientation: str = "0",
    ):
        """DingTalk 整体跳转 actionCard 消息"""
        return cls(
            type="actionCard",
            data={
                "title": title,
                "text": text,
                "singleTitle": single_title,
                "singleURL": single_url,
                "btnOrientation": btn_orientation,
            },
        )

    @classmethod
    def action_card_multi_btns(
        cls, title: str, text: str, btns: List[Any], btn_orientation: str = "0"
    ):
        """DingTalk 独立跳转 actionCard 消息"""
        return cls(
            type="actionCard",
            data={
                "title": title,
                "text": text,
                "btns": btns,
                "btnOrientation": btn_orientation,
            },
        )

    @classmethod
    def feed_card(cls, links: List[Any]):
        """DingTalk feedCard 消息"""
        return cls(type="feedCard", data={"links": links})

    @classmethod
    def at(
        cls,
        at_mobiles: Optional[List[str]] = None,
        at_user_ids: Optional[List[str]] = None,
        is_at_all: bool = False,
    ):
        """DingTalk At 信息"""
        return cls(
            type="at",
            data={
                "atMobiles": at_mobiles,
                "atUserIds": at_user_ids,
                "isAtAll": is_at_all,
            },
        )



"""DingTalk 适配器配置。"""
from alicebot.config import ConfigModel


class Config(ConfigModel):
    """DingTalk 配置类，将在适配器被加载时被混入到机器人主配置中。

    Attributes:
        host: 本机域名。
        port: 监听的端口。
        url: 路径。
        api_timeout: 进行 API 调用时等待返回响应的超时时间。
        app_secret: 机器人的 `appSecret`。
    """

    __config_name__ = "dingtalk"
    host: str = "127.0.0.1"
    port: int = 8080
    url: str = "/dingtalk"
    api_timeout: int = 1000
    app_secret: str = ""



"""DingTalk 协议适配器。

本适配器适配了钉钉企业自建机器人协议。
协议详情请参考: [钉钉开放平台](https://open.dingtalk.com/document/robots/robot-overview)。
"""
import base64
import hashlib
import hmac
import time
from typing import Any, Dict, Literal, Union

import aiohttp
from aiohttp import web

from alicebot.adapter import Adapter
from alicebot.log import error_or_exception, logger

from .config import Config
from .event import DingTalkEvent
from .exceptions import NetworkError
from .message import DingTalkMessage

__all__ = ["DingTalkAdapter"]


class DingTalkAdapter(Adapter[DingTalkEvent, Config]):
    """钉钉协议适配器。"""

    name: str = "dingtalk"
    Config = Config

    app: web.Application
    runner: web.AppRunner
    site: web.TCPSite

    session: aiohttp.ClientSession

    async def startup(self):
        """创建 aiohttp Application。"""
        self.app = web.Application()
        self.app.add_routes([web.post(self.config.url, self.handler)])

        self.session = aiohttp.ClientSession()

    async def run(self):
        """运行 aiohttp 服务器。"""
        self.runner = web.AppRunner(self.app)
        await self.runner.setup()
        self.site = web.TCPSite(self.runner, self.config.host, self.config.port)
        await self.site.start()

    async def shutdown(self):
        """清理 aiohttp AppRunner。"""
        await self.session.close()
        await self.site.stop()
        await self.runner.cleanup()

    async def handler(self, request: web.Request):
        """处理 aiohttp 服务器的接收。

        Args:
            request: aiohttp 服务器的 `Request` 对象。
        """
        if "timestamp" not in request.headers or "sign" not in request.headers:
            logger.error("Illegal http header, incomplete http header")
        elif abs(int(request.headers["timestamp"]) - time.time() * 1000) > 3600000:
            logger.error(
                f'Illegal http header, timestamp: {request.headers["timestamp"]}'
            )
        elif request.headers["sign"] != self.get_sign(request.headers["timestamp"]):
            logger.error(f'Illegal http header, sign: {request.headers["sign"]}')
        else:
            try:
                dingtalk_event = DingTalkEvent(adapter=self, **(await request.json()))
            except Exception as e:
                error_or_exception(
                    "Request parsing error:",
                    e,
                    self.bot.config.bot.log.verbose_exception,
                )
                return web.Response()
            await self.handle_event(dingtalk_event)
        return web.Response()

    def get_sign(self, timestamp: str) -> str:
        """计算签名。

        Args:
            timestamp: 时间戳。

        Returns:
            签名。
        """
        hmac_code = hmac.new(
            self.config.app_secret.encode("utf-8"),
            f"{timestamp}\n{self.config.app_secret}".encode(),
            digestmod=hashlib.sha256,
        ).digest()
        return base64.b64encode(hmac_code).decode("utf-8")

    async def send(
        self,
        webhook: str,
        conversation_type: Literal["1", "2"],
        msg: Union[str, Dict[str, Any], DingTalkMessage],
        at: Union[None, Dict[str, Any], DingTalkMessage] = None,
    ) -> Dict[str, Any]:
        """发送消息。

        Args:
            webhook: Webhook 网址。
            conversation_type: 聊天类型，"1" 表示单聊，"2" 表示群聊。
            msg: 消息。
            at: At 对象，仅在群聊时生效，默认为空。

        Returns:
            钉钉服务器的响应。

        Raises:
            TypeError: 传入参数类型错误。
            ValueError: 传入参数值错误。
            NetworkError: 调用 Webhook 地址时网络错误。
        """
        if isinstance(msg, DingTalkMessage):
            pass
        elif isinstance(msg, dict):
            msg = DingTalkMessage.raw(msg)
        elif isinstance(msg, str):
            msg = DingTalkMessage.text(msg)
        else:
            raise TypeError(
                f"msg must be str, Dict or DingTalkMessage, not {type(msg)!r}"
            )

        if at is not None:
            if isinstance(at, DingTalkMessage):
                if at.type == "at":
                    pass
                else:
                    raise ValueError(f'at.type must be "at", not {at.type}')
            elif isinstance(at, dict):
                at = DingTalkMessage.raw(at)
            else:
                raise TypeError(f"at must be Dict or DingTalkMessage, not {type(at)!r}")

        if conversation_type == "1":
            data = msg
        elif conversation_type == "2":
            if at is None:
                data = {"msgtype": msg.type, **msg.as_dict()}
            else:
                data = {"msgtype": msg.type, **msg.as_dict(), **at.as_dict()}
        else:
            raise ValueError(
                f'conversation_type must be "1" or "2" not {conversation_type}'
            )

        try:
            async with self.session.post(webhook, json=data) as resp:
                return await resp.json()
        except aiohttp.ClientError as e:
            raise NetworkError from e
